import unittest
from mock import MagicMock, patch

from celery import Celery

from cadasta.workertoolbox import setup


def mock_setup_func(success=True):
    if success:
        return MagicMock(return_value=True)
    return MagicMock(
        side_effect=AttributeError("Uh oh!"), __name__='SetupFuncB')


class TestSetup(unittest.TestCase):
    @patch('cadasta.workertoolbox.setup.SETUP_FUNCS',
           (mock_setup_func(True), mock_setup_func(False)))
    @patch('cadasta.workertoolbox.setup.logger')
    def test_caught_failures(self, logger):
        app = Celery()
        setup.setup_app(app, throw=False)

        from cadasta.workertoolbox.setup import SETUP_FUNCS
        for func in SETUP_FUNCS:
            func.assert_called_once_with(app)

        logger.exception.assert_called_once_with(
            'Failed to run setup function %r(app)', 'SetupFuncB')

        self.assertFalse(app.is_set_up)

    @patch('cadasta.workertoolbox.setup.SETUP_FUNCS',
           (mock_setup_func(True), mock_setup_func(False)))
    @patch('cadasta.workertoolbox.setup.logger')
    def test_thrown_failures(self, logger):
        app = Celery()
        with self.assertRaises(AttributeError):
            setup.setup_app(app, throw=True)

        from cadasta.workertoolbox.setup import SETUP_FUNCS
        for func in SETUP_FUNCS:
            SETUP_FUNCS[0].assert_called_once_with(app)

        self.assertFalse(logger.exception.called)
        self.assertFalse(app.is_set_up)

    @patch('cadasta.workertoolbox.setup.SETUP_FUNCS',
           (mock_setup_func(True), mock_setup_func(True)))
    @patch('cadasta.workertoolbox.setup.logger')
    def test_no_failures(self, logger):
        app = Celery()
        setup.setup_app(app, throw=True)

        from cadasta.workertoolbox.setup import SETUP_FUNCS
        for func in SETUP_FUNCS:
            SETUP_FUNCS[0].assert_called_once_with(app)

        self.assertFalse(logger.exception.called)
        self.assertTrue(app.is_set_up)



from ast import literal_eval
from os import environ as env
import pprint
import logging
import logging.config

from kombu import Exchange, Queue
from raven import Client
from raven.contrib.celery import register_signal, register_logger_signal

# Ensure signals are imported before app starts
from .signals import *  # NOQA
from . import DEFAULT_QUEUES


DEFAULT_LOGGING_FMT = '[%(asctime)s: %(levelname)s/%(processName)s %(message)s'
DEFAULT_LOGGING_CONFIG = {
    'version': 1,
    'formatters': {
        'colored': {
            'format': DEFAULT_LOGGING_FMT,
            'class': 'cadasta.workertoolbox.utils.ColorFormatter'
        },
        'default': {
            'format': DEFAULT_LOGGING_FMT,
        },
    },
    'handlers': {
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'colored',
        },
        'info_file': {
            'level': 'INFO',
            'class': 'logging.handlers.RotatingFileHandler',
            'formatter': 'default',
            'filename': 'app.info.log',
            'backupCount': 2,
            'maxBytes': 1024 * 1024 * 5,
        },
        'error_file': {
            'level': 'ERROR',
            'class': 'logging.handlers.RotatingFileHandler',
            'formatter': 'default',
            'filename': 'app.error.log',
            'backupCount': 2,
            'maxBytes': 1024 * 1024 * 5,
        },
    },
    'loggers': {
        '': {
            'handlers': ['console', 'info_file', 'error_file'],
        },
    }
}


class Config:
    def __init__(self, **kw):
        """
        Object to manage Celery application configuration.
        """
        self.ENV_PREFIX = kw.get(
            'ENV_PREFIX', env.get('CELERY_ENV_PREFIX', 'CELERY_'))

        # Assign any keyword argument to object
        for k, v in kw.items():
            setattr(self, k, v)

        # Configure Broker
        self.set('QUEUE_PREFIX', 'dev')
        self.set('broker_transport', 'sqs')
        self.set('broker_transport_options', {
                'region': 'us-west-2',
                'queue_name_prefix': '{}-'.format(self.QUEUE_PREFIX),

                # Ensure long-polling for SQS messages
                'wait_time_seconds': 20,
                # Wait up to 20 seconds for msg ack from worker (if
                # creating queue)
                'visibility_timeout': 20,
                # Retry immediately if cannot connect to SQS once
                'interval_start': 0,
                # Ensure error is raised if cannot connect to SQS twice
                'max_retries': 1,
            })

        # Setup Logging
        self.set('task_track_started', True)
        if self.set('SETUP_FILE_LOGGING', False):
            self.setup_file_logging()

        if self.set('SETUP_SENTRY_LOGGING', bool(env.get('SENTRY_DSN'))):
            assert env.get('SENTRY_DSN'), (
                'Required env variable for Sentry logging is not set')
            self.setup_sentry_logging()

        # Configure Result Backend
        self.set('RESULT_DB_USER', 'worker')
        self.set('RESULT_DB_PASS', 'cadasta')
        self.set('RESULT_DB_HOST', 'localhost')
        self.set('RESULT_DB_NAME', 'cadasta')
        self.set('RESULT_DB_PORT', '5432')
        self.set('result_backend', (
            'db+postgresql://{0.RESULT_DB_USER}:{0.RESULT_DB_PASS}@'
            '{0.RESULT_DB_HOST}:{0.RESULT_DB_PORT}/{0.RESULT_DB_NAME}'))
        try:
            self.result_backend = self.result_backend.format(self)
        except ValueError:
            raise ValueError(
                "Unable to render 'result_backend' value: %r" %
                self.result_backend)

        # Configure Routes & Exchanges
        self.set('task_default_exchange', 'task_exchange')
        self.set('task_default_exchange_type', 'topic')
        self.set('task_routes', self._route_task, from_env=False)

        # Configure Queues
        self.set('QUEUES', DEFAULT_QUEUES)
        self.set('PLATFORM_QUEUE_NAME', 'platform.fifo')
        if not hasattr(self, 'task_queues'):
            self.task_queues = self._generate_queues(
                self.QUEUES, self._default_exchange_obj,
                self.PLATFORM_QUEUE_NAME)

        # Configure Tasks
        self.set('imports', ('app.tasks',))
        self.set('CHORD_UNLOCK_MAX_RETRIES', 60 * 60 * 6)  # 6 hrs

        # Assign any other matching env variables to object
        for k, v in env.items():
            if not k.startswith(self.ENV_PREFIX):
                continue
            key = k.split(self.ENV_PREFIX, 1)[1].lower()
            if hasattr(self, key.lower()) or hasattr(self, key.upper()):
                continue
            if key in ('log_level', 'log_file'):  # Ignore Celery-set env vars
                continue
            setattr(self, key, v)

    def __repr__(self):
        attr_str = pprint.pformat(self.to_dict())
        return '{0.__class__.__name__}({1})'.format(self, attr_str)

    def to_dict(self):
        return {
            k: getattr(self, k)
            for k in dir(self)
            if (k.islower() and not k.startswith('_') and
                not callable(getattr(self, k)))
        }

    def set(self, keyword, default, from_env=True):
        """
        Set value on self if not already set. If unset, attempt to
        retrieve from environment variable of same name (unless disabled
        via 'from_env'). If 'default' value is not a string, evaluate
        environment variable as a Python type. If no env variables are
        found, fallback to 'default' value.
        """
        env_key = '{}{}'.format(self.ENV_PREFIX, keyword.upper())
        if hasattr(self, keyword):
            return getattr(self, keyword)
        value = default
        if from_env and (env_key in env):
            env_val = env.get(env_key)
            should_eval = not isinstance(default, str)
            try:
                value = literal_eval(env_val) if should_eval else env_val
            except (ValueError, SyntaxError):
                raise ValueError("Unable to cast %r to %r" % (
                    env_val, type.__name__))
        setattr(self, keyword, value)
        return getattr(self, keyword)

    def setup_file_logging(self, config=DEFAULT_LOGGING_CONFIG):
        self.set('worker_hijack_root_logger', False)
        logging.config.dictConfig(config)

    def setup_sentry_logging(self, sentry_client=None, level=logging.ERROR):
        self._sentry_client = (
            sentry_client or
            Client(**{k: v for k, v in {
                'dsn': env['SENTRY_DSN'],
                'name': env.get('SENTRY_NAME'),
                'environment': env.get('SENTRY_ENVIRONMENT'),
                'release': env.get('SENTRY_RELEASE'),
            }.items() if v})
        )
        register_logger_signal(self._sentry_client, loglevel=level)
        register_signal(self._sentry_client)

    @property
    def _default_exchange_obj(self):
        return Exchange(
            self.task_default_exchange,
            self.task_default_exchange_type)

    @staticmethod
    def _generate_queues(queues, exchange, platform_queue):
        """ Queues known by this worker """
        return set([
            Queue('celery', exchange, routing_key='celery'),
            Queue(platform_queue, exchange, routing_key='#'),
        ] + [
            Queue(q_name, exchange, routing_key=q_name)
            for q_name in queues
        ])

    def _route_task(self, name, args, kwargs, options, task=None, **kw):
        return {
            'routing_key': name.split('.')[0],
            'exchange': self._default_exchange_obj
        }



import logging
import unittest
from mock import patch

from cadasta.workertoolbox.conf import Config


class TestConfigClass(unittest.TestCase):

    @patch('cadasta.workertoolbox.conf.env', {'CELERY_FOO': '9876'})
    def test_set_env_vars(self):
        """
        Ensure that sets any CELERY_* env var to object
        """
        conf = Config()
        self.assertEqual(conf.foo, '9876')

    @patch('cadasta.workertoolbox.conf.env', {'CELERY_FOO': '9876'})
    def test_set_no_overwrite_uppercase(self):
        """
        Ensure that any CELERY_* env var won't be set to object if
        uppercase version already exists on object.
        """
        conf = Config(FOO='1234')
        self.assertFalse(hasattr(conf, 'foo'))
        self.assertEqual(conf.FOO, '1234')

    @patch('cadasta.workertoolbox.conf.env', {'CELERY_FOO': '9876'})
    def test_set_no_overwrite(self):
        """
        Ensure that sets to provide init keyword argument instead of env
        variable or function 'default' argument
        """
        conf = Config(foo=1234)
        conf.set('foo', 5678)
        self.assertEqual(conf.foo, 1234)

    @patch('cadasta.workertoolbox.conf.env', {})
    def test_set_uses_default(self):
        """
        Ensure uses function 'default' argument if no env or preset value
        """
        conf = Config()
        conf.set('foo', 1234)
        self.assertEqual(conf.foo, 1234)

    def test_set_uses_env(self):
        """
        Ensure that if the 'default' is a string, won't evaluate env variable
        """
        conf = Config()
        with patch('cadasta.workertoolbox.conf.env', {'CELERY_FOO': 'False'}):
            conf.set('foo', True)
        self.assertEqual(conf.foo, False)

    def test_set_no_eval(self):
        """
        Ensure that if the 'default' is a string, won't evaluate env variable
        """
        conf = Config()
        with patch('cadasta.workertoolbox.conf.env', {'CELERY_FOO': 'False'}):
            conf.set('foo', 'True')
        self.assertEqual(conf.foo, 'False')

    def test_set_bad_env_val(self):
        """
        Ensure error is thrown if the 'default' is a string, and env
        variable is not evaluatable value
        """
        conf = Config()
        with patch('cadasta.workertoolbox.conf.env', {'CELERY_FOO': ''}):
            with self.assertRaises(ValueError):
                conf.set('foo', True)

    def test_override(self):
        """ Ensure default params can be overridden """
        conf = Config(broker_transport='foo')
        self.assertEqual(conf.broker_transport, 'foo')

    def test_repr(self):
        conf = Config(a=1, b=2, z=3)
        self.assertTrue(repr(conf).startswith("Config({'a': 1,\n 'b': 2,\n"))
        self.assertTrue(repr(conf).endswith(" 'z': 3})"))

    def test_queues(self):
        self.assertEqual(
            Config(QUEUES=['foo']).QUEUES,
            ['foo'])

    def test_backend_default(self):
        self.assertEqual(
            Config().result_backend,
            'db+postgresql://worker:cadasta@localhost:5432/cadasta')

    def test_backend_custom(self):
        self.assertEqual(
            Config(result_backend=':memory:').result_backend,
            ':memory:')

    def test_backend_custom_rendered(self):
        self.assertEqual(
            Config(
                result_backend='{0.foo}:{0.BAR}',
                foo='abc', BAR='def'
            ).result_backend,
            'abc:def'
        )

    def test_backend_custom_error(self):
        with self.assertRaises(ValueError) as context:
            Config(
                result_backend='{0.foo}:{}',
                foo='abc'
            )
        self.assertEqual(
            getattr(context.exception, 'message', context.exception.args[0]),
            "Unable to render 'result_backend' value: '{0.foo}:{}'"
        )

    def test_default_chord_unlock_max_retries(self):
        conf = Config()
        self.assertTrue(isinstance(conf.CHORD_UNLOCK_MAX_RETRIES, int))

    @patch('cadasta.workertoolbox.conf.Config.setup_file_logging')
    def test_default_no_setup_file_logging(self, setup_file_logging):
        Config()
        self.assertFalse(setup_file_logging.called)

    @patch('cadasta.workertoolbox.conf.Config.setup_file_logging')
    def test_setup_file_logging_argument(self, setup_file_logging):
        Config(SETUP_FILE_LOGGING=True)
        setup_file_logging.assert_called_once_with()

    @patch('cadasta.workertoolbox.conf.logging')
    def test_setup_file_logging(self, logging):
        my_logging_config = {}
        Config(SETUP_FILE_LOGGING=False).setup_file_logging(my_logging_config)
        logging.config.dictConfig.assert_called_once_with(my_logging_config)

    @patch('cadasta.workertoolbox.conf.env', {
        'SENTRY_DSN': 'https://example.com',
        'SENTRY_NAME': 'foo',
        'SENTRY_ENVIRONMENT': 'bar',
        'SENTRY_RELEASE': '987',
    })
    @patch('cadasta.workertoolbox.conf.Client')
    @patch('cadasta.workertoolbox.conf.register_logger_signal')
    @patch('cadasta.workertoolbox.conf.register_signal')
    def test_setup_sentry_tools(self, register_signal, logger_signal, Client):
        """ Ensure sentry logging is called if env variable is set """
        Config()
        Client.assert_called_once_with(
            dsn='https://example.com',
            name='foo',
            environment='bar',
            release='987',
        )
        logger_signal.assert_called_once_with(Client.return_value, loglevel=logging.ERROR)
        register_signal.assert_called_once_with(Client.return_value)

    @patch('cadasta.workertoolbox.conf.env', {'SENTRY_DSN': 'https://example.com'})
    @patch('cadasta.workertoolbox.conf.Client')
    @patch('cadasta.workertoolbox.conf.register_logger_signal')
    @patch('cadasta.workertoolbox.conf.register_signal')
    def test_setup_sentry_tools_override(self, register_signal, logger_signal, Client):
        """
        Ensure sentry logging is not called if env variable is set but setup
        is set to false
        """
        Config(SETUP_SENTRY_LOGGING=False)
        self.assertFalse(Client.called)
        self.assertFalse(register_signal.called)
        self.assertFalse(logger_signal.called)

    @patch('cadasta.workertoolbox.conf.env', {})
    @patch('cadasta.workertoolbox.conf.Client')
    @patch('cadasta.workertoolbox.conf.register_logger_signal')
    @patch('cadasta.workertoolbox.conf.register_signal')
    def test_setup_sentry_tools_not_called(self, register_signal, logger_signal, Client):
        """
        Ensure sentry logging is not called if sentry env variabels are unset
        """
        Config()
        self.assertFalse(Client.called)
        self.assertFalse(register_signal.called)
        self.assertFalse(logger_signal.called)
